package NeuralNetwork;

/**
 * Created by Maaarty on 7.3.2017.
 */

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


/**
 * <p>
 * Utility class for working with the MNIST database.
 * <p>
 * Provides methods for traversing the images and labels data files separately,
 * as well as simultaneously.
 * <p>
 * Provides also method for exporting an image by writing it as a PPM file.
 * <p>
 * Example usage:
 * <pre>
 *  MnistManager m = new MnistManager("t10k-images.idx3-ubyte", "t10k-labels.idx1-ubyte");
 *  m.setCurrent(10); //index of the image that we are interested in
 *  int[][] image = m.readImage();
 *  System.out.println("Label:" + m.readLabel());
 *  MnistManager.writeImageToPpm(image, "10.ppm");
 * </pre>
 */
public class GomokuManagerSmall {
    GomokuImageFileSmall images;
    private GomokuLabelFileSmall labels;
    private float divider;

    private ArrayList<int[]> imagesArr;
    private ArrayList<float[]> labelsArr;
    private static final int HEADER_SIZE = 8;

    /**
     * Writes the given image in the given file using the PPM data format.
     *
     * @param image
     * @param ppmFileName
     * @throws IOException
     */
    public static void writeImageToPpm(int[][] image, String ppmFileName) throws IOException {
        try (BufferedWriter ppmOut = new BufferedWriter(new FileWriter(ppmFileName))) {
            int rows = image.length;
            int cols = image[0].length;
            ppmOut.write("P3\n");
            ppmOut.write("" + rows + " " + cols + " 255\n");
            for (int i = 0; i < rows; i++) {
                StringBuilder s = new StringBuilder();
                for (int j = 0; j < cols; j++) {
                    s.append(image[i][j] + " " + image[i][j] + " " + image[i][j] + "  ");
                }
                ppmOut.write(s.toString());
            }
        }

    }

    /**
     * Constructs an instance managing the two given data files. Supports
     * <code>NULL</code> value for one of the arguments in case reading only one
     * of the files (images and labels) is required.
     *
     * @param imagesFile Can be <code>NULL</code>. In that case all future operations
     *                   using that file will fail.
     * @param labelsFile Can be <code>NULL</code>. In that case all future operations
     *                   using that file will fail.
     * @throws IOException
     */
    public GomokuManagerSmall(String imagesFile, String labelsFile, boolean train) throws IOException {
        divider = 1;
        if (labelsFile != null) {
            labels = new GomokuLabelFileSmall(labelsFile, "r");
            if (train) labelsArr = labels.readLabels(GomokuDataFetcherSmall.NUM_EXAMPLES);
            else labelsArr = labels.readLabels(GomokuDataFetcherSmall.NUM_EXAMPLES_TEST);
        }

        if (imagesFile != null) {
            images = new GomokuImageFileSmall(imagesFile, "r");
            if (train) imagesArr = images.readImagesUnsafe(GomokuDataFetcherSmall.NUM_EXAMPLES, labelsArr);
            else imagesArr = images.readImagesUnsafe(GomokuDataFetcherSmall.NUM_EXAMPLES_TEST, labelsArr);
        }

        System.out.println();
    }

    public GomokuManagerSmall(String imagesFile, String labelsFile) throws IOException {
        this(imagesFile, labelsFile, true);
    }

    /**
     * Reads the current image.
     *
     * @return matrix
     * @throws IOException
     */
    public int[][] readImage() throws IOException {
        if (images == null) {
            throw new IllegalStateException("Images file not initialized.");
        }
        return images.readImage();
    }

    public float[] readImageUnsafe(int i) {
        float[] newExample = new float[imagesArr.get(i).length];
        for (int j = 0; j < imagesArr.get(i).length; j++) {
            newExample[j] = imagesArr.get(i)[j] / divider;
        }
        return newExample;
    }

    /**
     * Set the position to be read.
     *
     * @param index
     */
    public void setCurrent(int index) {
        images.setCurrentIndex(index);
        labels.setCurrentIndex(index);
    }

    /**
     * Reads the current label.
     *
     * @return int
     * @throws IOException
     */
    public int readLabel() throws IOException {
        if (labels == null) {
            throw new IllegalStateException("labels file not initialized.");
        }
        return labels.readLabel();
    }

    public float[] readLabel(int i) {
        float[] newExample = new float[labelsArr.get(i).length];
        for (int j = 0; j < labelsArr.get(i).length; j++) {
            newExample[j] = labelsArr.get(i)[j] / divider;
        }
        return newExample;
    }

    /**
     * Get the underlying images file as {@link GomokuImageFile}.
     *
     * @return {@link GomokuImageFile}.
     */
    public GomokuImageFileSmall getImages() {
        return images;
    }

    /**
     * Get the underlying labels file as {@link GomokuLabelFile}.
     *
     * @return {@link GomokuLabelFile}.
     */
    public GomokuLabelFileSmall getLabels() {
        return labels;
    }

    /**
     * Close any resources opened by the manager.
     */
    public void close() {
        if (images != null) {
            try {
                images.close();
            } catch (IOException e) {
            }
            images = null;
        }
        if (labels != null) {
            try {
                labels.close();
            } catch (IOException e) {
            }
            labels = null;
        }
    }

    public ArrayList<float[]> getLabelsArr() {
        return labelsArr;
    }
}
